% CLASS: ModelParams
% 
% HIERARCHY (Enumeration of the various super- and subclasses)
%   Superclasses: hgsetget (MATLAB handle class)
%   Subclasses: N/A
% 
% TYPE (Abstract or Concrete)
%   Concrete
%
% DESCRIPTION (High-level overview of the class)
%   This class is used by the main AMP-MMV inference functions, (see
%   sp_mmv_wrapper_fxn.m, sp_mmv_fxn.m, sp_mmv_serial_fxn.m), in order to
%   specify the parameters (or initialization of the parameters if
%   performing EM parameter learning) that make up the signal/measurement
%   model.  The parameters that are needed to define the model are
%   described briefly below (see PROPERTIES), and a full description of the
%   signal/measurement model can be found in either sp_mmv_fxn.m or
%   sp_mmv_serial_fxn.m.
%
%   To create a ModelParams object, there are two constructors to choose
%   from (see METHODS section below).  The first constructor simply accepts
%   an input for each of the required model parameters, i.e.,
%   ModelParams(lambda, zeta, sigma2, alpha, sig2e).  The second
%   constructor accepts a SigGenParams object, along with sig2e, and copies
%   over the required parameter values, i.e., ModelParams(SigGenParams,
%   sig2e).
%
% PROPERTIES (Model Parameters)
%   lambda          Scalar activity probability for s_n, i.e., 
%                   Pr{s_n = 1} = lambda
%   zeta        	(Complex) scalar of theta means, i.e.,
%                   E[theta_n(t)] = zeta
%   sigma2          Scalar steady-state theta variances, i.e.
%                   var{theta_n(t)} = sigma2
%   alpha           Scalar Gauss-Markov "innovation rate" parameter, btwn.
%                   0 and 1 (Note: Correlation coeff. = 1 - alpha)
%   sig2e           Variance of the additive Gaussian measurement noise
%
%   ***Note: The parameters lambda, zeta, sigma2, and alpha can, in
%            addition to scalar inputs, accept N-by-1 vector inputs if you
%            wish to assign model parameters on a per-coefficient basis.
%
% METHODS (Subroutines/functions)
%   ModelParams(lambda, zeta, sigma2, alpha, sig2e)
%       - Primary constructor for creating a ModelParams object
%   ModelParams(SigGenParams, sig2e)
%       - Secondary constructor that uses the SigGenParams object that
%         generated the data, in addition to sig2e, to fill in the values
%         of the model parameters
%   set(obj, 'ParameterName', Value)
%       - Method that can be used to set the value of a parameter once the
%         object of class ModelParams, obj, has been constructed
%   print()
%       - Print the current value of each property to the command window
%                   
%

%
% Coded by: Justin Ziniel, The Ohio State Univ.
% E-mail: zinielj@ece.osu.edu
% Last change: 12/29/11
% Change summary: 
%       - Created from ModelParams v1.0 (12/29/11; JAZ)
% Version 1.1
%

classdef ModelParams < hgsetget

    properties
        lambda;  	% Pr{s_n = 1}
        zeta;     	% E[theta_n(t)]
        sigma2;   	% var{theta_n(t)}
        alpha;    	% Amplitude process innovation rate
        sig2e;      % AWGN variance
    end % properties
    
    properties (Dependent)
        rho;      	% The variance of the amplitude driving noise
    end
    
    methods
        % Constructor
        function obj = ModelParams(varargin)
            if nargin == 5
                obj.lambda = varargin{1};
                obj.zeta = varargin{2};
                obj.sigma2 = varargin{3};
                obj.alpha = varargin{4};
                obj.sig2e = varargin{5};
            elseif nargin == 2
                if strcmp('SigGenParams', class(varargin{1}))
                    obj.lambda = varargin{1}.lambda;
                    obj.zeta = varargin{1}.zeta;
                    obj.sigma2 = varargin{1}.sigma2;
                    obj.alpha = varargin{1}.alpha;
                    obj.sig2e = varargin{2};
                else
                    error(['First argument to ModelParams constructor is ' ...
                        'not an object of the SigGenParams class'])
                end
            else
                error('Incorrect number of arguments to ModelParams constructor')
            end
        end
        
        % Set method for lambda
        function obj = set.lambda(obj, lambda)
           if lambda >= 0 & lambda <= 1
               obj.lambda = lambda;
           else
              error('Invalid assignment: lambda')
           end
        end
        
        % Set method for zeta
        function obj = set.zeta(obj, zeta)
            obj.zeta = zeta;
           if ~isreal(zeta)
               fprintf('Complex mean detected - Setting version = ''C''\n')
               obj.version = 'C';
           end
        end
        
        % Set method for sigma2
        function obj = set.sigma2(obj, sigma2)
           if sigma2 > 0
               obj.sigma2 = sigma2;
           else
              error('Invalid assignment: sigma2')
           end
        end
        
        % Set method for alpha
        function obj = set.alpha(obj, alpha)
           if alpha >= 0 & alpha < 1
               obj.alpha = alpha;
           else
              error('Invalid assignment: alpha')
           end
        end
        
        % Set method for SNRmdB
        function obj = set.sig2e(obj, sig2e)
           if numel(sig2e) == 1 && sig2e >= 0
               obj.sig2e = sig2e;
           else
              error('Invalid assignment: sig2e')
           end
        end
        
        % Get method for dependent property rho (which is set based on the
        % values of alpha and sigma2 to ensure a stationary amplitude
        % variance over time
        function rho = get.rho(obj)
            if numel(obj.alpha) == numel(obj.sigma2)
                rho = (2 - obj.alpha) .* obj.sigma2 ./ obj.alpha;
            elseif numel(obj.alpha) > 1 && numel(obj.sigma2) == 1
                rho = obj.sigma2 * (2 - obj.alpha) ./ obj.alpha;
            elseif numel(obj.sigma2) > 1 && numel(obj.alpha) ==  1
                rho = ((2 - obj.alpha)/obj.alpha)*obj.sigma2;
            else
                error('Dimension mismatch between alpha and sigma2')
            end
        end
        
        % Get method for grabbing all the properties at once
        function [lambda, zeta, sigma2, alpha, sig2e, rho] = getParams(obj)
            lambda = obj.lambda;
            zeta = obj.zeta;
            sigma2 = obj.sigma2;
            alpha = obj.alpha;
            sig2e = obj.sig2e;
            rho = obj.rho;
        end
        
        % Print the current configuration to the command window
        function print(obj)
            fprintf('****************************************\n')
            fprintf('        Signal Model Parameters\n')
            fprintf('****************************************\n')
            fprintf('Activity probability (lambda): %s\n', form(obj, obj.lambda))
            fprintf('           Active mean (zeta): %s\n', form(obj, obj.zeta))
            fprintf('     Active variance (sigma2): %s\n', form(obj, obj.sigma2))
            fprintf('      Innovation rate (alpha): %s\n', form(obj, obj.alpha))
        	fprintf('        AWGN variance (sig2e): %s\n', form(obj, obj.sig2e))
        end
          
    end % methods
    
    methods (Access = private)
        % This method is called by the print method in order to format how
        % properties are printed.  Properties that are scalars will have
        % their values printed to the command window, while arrays will
        % have their dimensions, minimal, and maximal values printed to the
        % screen
        function string = form(obj, prop)
            if numel(prop) == 1
                string = num2str(prop);
            else
                string = sprintf('%d-by-%d array (Min: %g, Max: %g)', ...
                    size(prop, 1), size(prop, 2), min(min(prop)), ...
                    max(max(prop)));
            end
        end
    end % Private methods
   
end % classdef