%% AMP-MMV Demo
% This demo file will provide several examples of how to use the various
% tools that make up the AMP-MMV software package.  We realize that there
% is a lot to read in this demo, but taking the time to carefully follow
% the examples we've provided should provide the reader with a fairly
% thorough understanding of how to correctly operate AMP-MMV.  The file
% tutorial.pdf in the top-level directory of this software package provides
% a more visually appealing version of this demo.
%
% Note that in order for this demo script to work properly, all of the
% folders in the AMP-MMV software package must be added to MATLAB's path.
% This can be accomplished easily by running the script add2path.m, which
% can be found in the top-level directory of this software package.

%%%
% Coded by: Justin Ziniel, The Ohio State Univ.
% E-mail: zinielj @ ece.osu.edu
% Last change: 12/29/11
% Change summary: 
%       - Created from demo v1.0 (12/29/11; JAZ)
% Version 1.1
%

% Initialize the random number stream
RandStream.setDefaultStream(RandStream('mt19937ar', 'Seed', 0));

%% Example 1: A Simple Recovery
%
% In this first example, we will demonstrate how to obtain a sample
% MMV recovery with just a few lines of code.
%
%
% First, let's generate some synthetic data according to AMP-MMV's signal
% model.  To do this, we will use the function |SIGNAL_GEN_FXN|, which
% requires as an input an object of the custom-defined class SigGenParams.
% (The class definition file can be found in the folder ClassDefs.)  There
% are several ways to create objects of this class.  The easiest is to use
% a constructor with no arguments, which will construct an object using all
% default values for the parameters of the signal model.

SigGenObj = SigGenParams();     % Call the default constructor

%%%
% Now we can look at the parameter values associated with our object by
% using its print method, which will print the current configuration to the
% command window.

SigGenObj.print();      % Print the current signal model configuration

%%
% By looking at the command window printout, we can see that the default
% signal model consists of a signal of length |N| = 1024, measurement vectors
% of length |M| = 256, and a total of |T| = 4 timesteps.  The measurement
% matrix will be randomly constructed with IID Gaussian elements.  The
% next several parameters specify the signal prior.  We see that the
% apriori probability that a given coefficient |x_n(t)| is non-zero is given
% by |lambda| = 0.08.  For those coefficients that are non-zero, they have a
% mean of |zeta| = 0, and a variance of |sigma2| = 1.  The active coefficient
% amplitudes evolve over time with a correlation equal to 1 - |alpha| = 1 -
% 0.10 = 0.90.  Finally, we see that the additive noise will have a
% variance chosen to yield an SNR of 25dB, and that the data generated will
% be complex-valued.
%
% Having constructed the object that contains our model parameters, we can
% now pass it to |SIGNAL_GEN_FXN|, which will produce 6 outputs...

[x_true, y, A, support, K, sig2e] = signal_gen_fxn(SigGenObj);

%%%
% ...the first of which, |x_true|, is a 1-by-T cell array, with cell t
% containing the value of the true length-N signal at timestep t, i.e.,
% |x_true{t}|.  The second output, |y|, is a 1-by-T cell array of length-M
% measurements.  The third output, |A|, is a 1-by-T cell array of M-by-N
% (identical) measurement matrices.  (We defer for now the reason for
% giving duplicates of the measurement matrix.)  The fourth output is a
% vector of indices of the non-zero coefficients of the true signal, and
% the fifth output, |K|, is the cardinality of the support (i.e., the 
% number of true non-zeroes).  The sixth and final output, |sig2e|, is the 
% variance of the AWGN corrupting noise.
%
% Now that we have a realization to work with, we can now use AMP-MMV to
% attempt to recover |x_true| from the noisy measurements, |y|.  The
% easiest way to do this is to call |SP_MMV_WRAPPER_FXN|, which will try
% several different configurations of AMP-MMV (e.g., different message
% passing schedules, different numbers of smoothing iterations, etc.) until
% it finds a solution with a residual energy that is small enough.
%
% In order to use |SP_MMV_WRAPPER_FXN|, we need at least |y|, |A|, and a
% third input, |Params|, which is an object of the class ModelParams, and
% contains parameters that specify the signal model.  Since we have
% generated data according to our signal model, (and for now, we will
% assume perfect knowledge of the model parameters), we can construct a
% |Params| object easily by passing our SigGenParams object, |SigGenObj|,
% to the ModelParams constructor, along with |sig2e|, and it will copy over
% the required parameters:

Params = ModelParams(SigGenObj, sig2e);     % Construct ModelParams object

%%
% Now let's verify that our Params object was constructed correctly by
% printing its property values to the command window:

Params.print();     % Print the value of the Params object properties

%%
% Everything looks good here, so we are now ready to call
% |SP_MMV_WRAPPER_FXN|.  Since we are not providing an optional fourth
% input of runtime configuration options, AMP-MMV will use a default
% runtime configuration.

% Obtain a recovery using AMP-MMV
[x_hat, v_hat, lambda_hat] = sp_mmv_wrapper_fxn(y, A, Params);

%%
% The first output, |x_hat|, is a 1-by-T cell array of MMSE signal
% estimates.  The second output, |v_hat|, contains the posterior marginal
% variances of each MMSE estimate in a 1-by-T cell array, i.e., 
% |v_hat{t}(n) = var{x_n(t) | y}.  The final output, |lambda_hat|, is a 
% length-N vector of posterior activity probabilities, i.e., 
% |lambda_hat(n) = p(s_n = 1 | y)|.
%
% In addition to returning these three outputs, |SP_MMV_WRAPPER_FXN| has
% printed some information to the command window.  We see that the first
% things that it printed were the runtime options that it was using (the
% defaults) during execution.  For example, we can see from the printout
% that the maximum number of smoothing iterations equaled the minimum
% number (5), and that during the AMP message passes that occurred within
% each frame, a maximum of 15 AMP iterations were allowed.  The EM
% parameter learning procedure is enabled by default, and the algorithm is
% set to operate verbosely.  We also see that it first attempted a parallel
% message passing schedule, and we can observe the EM parameter updates of
% each of the model parameters after each smoothing iteration.  Since the
% residual energy obtained from using this initial schedule was below a
% pre-defined threshold, the algorithm terminates.
%
% Now let's check the quality of the solution obtained using AMP-MMV.
% We'll use timestep-averaged normalized mean square error (TNMSE) and the
% normalized support error rate (NSER) as our performance metrics.  As a
% support estimate, we will use a simple ML estimator that relies on our
% vector of posterior activity probabilities, |lambda_hat|.

s_hat = (lambda_hat > 1/2);     % ML support estimate

T = SigGenObj.T;        % Total # of timesteps

TNMSE = sum(sum(([x_true{:}]-[x_hat{:}]).^2, 1)./sum([x_true{:}].^2, 1))/T;
NSER = nser(support, find(s_hat == 1));

fprintf('TNMSE: %g dB\n', 10*log10(TNMSE));
fprintf('NSER: %g\n', NSER);

%%
% From the printouts to the command window, we see that AMP-MMV performed
% well on this problem in both TNMSE and NSER.  To finish this example,
% let's visually observe the true signal and the recovered signal:

figure(1);
subplot(121); imagesc(abs([x_true{:}])); colorbar
xlabel('Timestep (t)'); ylabel('Index [n]'); title('True signal')
subplot(122); imagesc(abs([x_hat{:}])); colorbar
xlabel('Timestep (t)'); ylabel('Index [n]'); title('AMP-MMV recovery')



%% Example 2: Customizing AMP-MMV
%
% In the previous example, we used default values for all of the functions
% that we called.  In this example, we will illustrate how to choose
% non-default values for the parameters.
%
% As before, we will generate synthetic data according to our signal model
% by calling |SIGNAL_GEN_FXN|.  Once again, we start by constructing an
% object of the SigGenParams class.  Previously, we called the SigGenParams
% constructor with no arguments, which created an object of all default
% values.  Suppose now that we'd like to come up with a more challenging
% test case.  Let's keep the signal dimension, |N|, the same, but reduce
% the number of measurements at each timestep, |M|, to 128, giving an
% undersampling ratio of |N/M = 8|.  Let's likewise adjust the prior
% activity probability, |lambda|, so that |M/K = 2|, that is, |lambda =
% 0.0625|.  We can use the following constructor call:

SigGenObj = SigGenParams('M', 128, 'lambda', 0.0625);
SigGenObj.print();

%%
% From the printout to the command window, we can see that |SigGenObj| has
% used our two custom settings, and left everything else at their default
% values.  Suppose we realize now that we wanted to change the innovation
% rate, |alpha|, as well, to |alpha = 0.05|.  We could just construct
% |SigGenObj| all over again, and include our custom choice of alpha in the
% constructor call.  Alternatively, we can change the value of |alpha| for
% the object we just constructed:

SigGenObj.alpha = 0.05;
SigGenObj.print();

%%
% From the printout, we see that the value of |alpha| was successfully
% updated.
%
% Now, let's generate our synthetic data by calling |SIGNAL_GEN_FXN|:

[x_true, y, A, support, K, sig2e] = signal_gen_fxn(SigGenObj);

%%
% As before, in order to obtain a recovery with AMP-MMV, we need to provide
% it with an initialization of the model parameters.  Last time, we created
% a ModelParams object, |Params|, with the true parameter values as an
% initialization by using the constructor ModelParams(SigGenParams, sig2e).
% This time, let's start with an initialization that is a little bit off
% from the true values.  To do this, we will use an alternative
% constructor, which has the syntax ModelParams(lambda, zeta, sigma2, alpha, sig2e).  
% Here's how we'll initialize:

Params = ModelParams((1/4)*SigGenObj.lambda, 0, 2, 0.029, 1/2*sig2e);
Params.print();

%%
% In the last example, we used default runtime options for AMP-MMV.  This
% time, let's customize some of those options.  To do this, we need to
% create an object of the class Options, which we will then pass to
% AMP-MMV.  Since this is a more challenging example, let's increase the
% default maximum number of smoothing iterations to 10, and let's increase
% the maximum number of inner AMP iterations to 25.  We can use the
% following constructor:

RunOpt = Options('smooth_iters', 10, 'inner_iters', 25);
RunOpt.print();

%%
% As we can see from the command window printout, our custom options appear
% within the |RunOpt| object we just created.  Notice though that our
% constructed object has verbosity defaulting to silent.  Let's change this
% so that we can see what AMP-MMV is doing while it runs.

RunOpt.verbose = true;
RunOpt.print();

%%
% Now everything is ready for us to call AMP-MMV.  As before, we'll use
% |SP_MMV_WRAPPER_FXN| to allow us to try multiple different schedules,
% only this time we will provide it with our custom runtime options.

[x_hat, v_hat, lambda_hat] = sp_mmv_wrapper_fxn(y, A, Params, RunOpt);

%%
% Once again, we see that |SP_MMV_WRAPPER_FXN| first tried a parallel
% message passing schedule, however this time, the resultant residual
% energy was not low enough, given our initialization of sig2e, to allow
% AMP-MMV to terminate.  Therefore, it next attempted a serial message
% passing schedule, also with 10 smoothing iterations.  This time, the
% residual energy is low enough, and so |SP_MMV_WRAPPER_FXN| reports that
% it is returning the recovery from the serial message passing schedule.
%
% Once more, let's compute our performance metrics, and look at a plot of
% our recovery:

s_hat = (lambda_hat > 1/2);
TNMSE = sum(sum(([x_true{:}]-[x_hat{:}]).^2, 1)./sum([x_true{:}].^2, 1))/T;
NSER = nser(support, find(s_hat == 1));
fprintf('TNMSE: %g dB\n', 10*log10(TNMSE));
fprintf('NSER: %g\n', NSER);
figure(1);
subplot(121); imagesc(abs([x_true{:}])); colorbar
xlabel('Timestep (t)'); ylabel('Index [n]'); title('True signal')
subplot(122); imagesc(abs([x_hat{:}])); colorbar
xlabel('Timestep (t)'); ylabel('Index [n]'); title('AMP-MMV recovery')

%%
% With a TNMSE of -35 dB, and an NSER of 0, we can see that AMP-MMV
% performed quite well, despite the fact that we initialized it with
% inaccurate model parameters, which it refined during execution.



%% Example 3: Using a particular schedule, and running the SKS
%
% In previous experiments, we used |SP_MMV_WRAPPER_FXN| to allow us to
% cycle through multiple different configurations of AMP-MMV if needed,
% changing the schedule between serial and parallel, and adjusting the
% number of smoothing iterations.  In this example, we will illustrate how
% to obtain a recovery using AMP-MMV only once, by calling the function
% that implements the specific message passing schedule we wish to use.  We
% will also demonstrate how to obtain a lower bound on the achievable TNMSE
% performance by running the support-aware Kalman smoother.
%
% If we are only interested in the parallel message passing schedule, then
% we can call the function |SP_MMV_FXN|, which accepts the same arguments
% as |SP_MMV_WRAPPER_FXN|.  Alternatively, if we are interested in the
% serial message passing schedule, which likewise uses the same arguments
% as |SP_MMV_WRAPPER_FXN|, then the appropriate function call is
% |SP_MMV_SERIAL_FXN|.
%
% For this example, let's use the parallel message passing schedule.
% First let's generate a signal according to our signal model, with some 
% customizations of the default signal model.  Let's also try generating
% real-valued, rather than complex-valued data

SigGenObj = SigGenParams('M', 160, 'lambda', 0.0625, 'zeta', 1, ...
    'version', 'R');
SigGenObj.print();
[x_true, y, A, support, K, sig2e] = signal_gen_fxn(SigGenObj);

%%
% Now we need to create our ModelParams object.  Let's provide AMP-MMV with
% an inaccurate initialization once again.  Let's also create a custom
% Options object with an increased number of smoothing iterations.  We'll
% bump this maximum number of smoothing iterations way up so that we can
% observe the parameter estimates move towards their true values.

Params = ModelParams(1/5*SigGenObj.lambda, 0, 1, 0.50, 1e-3);
Params.print();

RunOpt = Options('smooth_iters', 70, 'verbose', true);
RunOpt.print();

%%
% Now we are ready to directly call the parallel message passing schedule
% implementation of AMP-MMV:

[x_hat, v_hat, lambda_hat] = sp_mmv_fxn(y, A, Params, RunOpt);

%%
% And let's compute our performance metrics on the recovery we obtained
% using this schedule, and plot the recovery:

s_hat = (lambda_hat > 1/2);
TNMSE = sum(sum(([x_true{:}]-[x_hat{:}]).^2, 1)./sum([x_true{:}].^2, 1))/T;
NSER = nser(support, find(s_hat == 1));
fprintf('TNMSE: %g dB\n', 10*log10(TNMSE));
fprintf('NSER: %g\n', NSER);
figure(1);
subplot(121); imagesc([x_true{:}]); colorbar
xlabel('Timestep (t)'); ylabel('Index [n]'); title('True signal')
subplot(122); imagesc([x_hat{:}]); colorbar
xlabel('Timestep (t)'); ylabel('Index [n]'); title('AMP-MMV recovery')


%%
% Now let's run the support-aware Kalman smoother so that we can see how
% well AMP-MMV performed compared to this genie-aided bound.  The function
% that we will call is |GENIE_MULTI_FRAME_FXN|.  In order to obtain a
% meaningful bound, we will provide the function with the true underlying
% model parameters.  Let's construct this ModelParams object by using the
% constructor that simply copies over the information from our SigGenParams
% object, |SigGenObj|:

SKSParams = ModelParams(SigGenObj, sig2e);
SKSParams.print();

%%
% We will also create a custom Options object that tells
% |GENIE_MULTI_FRAME_FXN| not to use the AMP algorithm to perform message
% passing within each frame, but rather use the slower (but theoretically
% grounded) exact belief propagation algorithm.

SKSRunOpt = Options('smooth_iters', 200, 'alg', 'BP');
SKSRunOpt.print();

%%
% Now, recalling that |support| is the variable that contains the true
% signal support, let's run the SKS:

x_sks = genie_multi_frame_fxn(y, A, support, SKSParams, SKSRunOpt);

%%
% Now let's see how AMP-MMV stacked up against the SKS by calculating the
% TNMSE of the SKS recovery (note that the NSER for the SKS will always be
% zero)

TNMSE = sum(sum(([x_true{:}]-[x_sks{:}]).^2, 1)./sum([x_true{:}].^2, 1))/T;
fprintf('TNMSE: %g dB\n', 10*log10(TNMSE));
figure(1);
subplot(121); imagesc([x_true{:}]); colorbar
xlabel('Timestep (t)'); ylabel('Index [n]'); title('True signal')
subplot(122); imagesc([x_sks{:}]); colorbar
xlabel('Timestep (t)'); ylabel('Index [n]'); title('SKS recovery')