% SP_MMV_FXN     Function for performing soft-input, soft-output sparse
% reconstruction of a time-series of complex-valued signals, {x(t)}, 
% t = 1, ..., T, with each x(t) being N-dimensional.  Complex-valued
% observations, {y(t)}, are obtained through a noisy linear combination,
% y(t) = A(t)x(t) + e(t), with y(t) being an M(t) dimensional vector of
% observations, A(t) being an M(t)-by-N complex-valued measurement
% matrix, and e(t) being circular, complex, additive Gaussian noise with
% covariance matrix sig2e*eye(M(t)).
%
% At any given time, t, it is assumed that the coefficients of x(t),
% x_n(t), can be represented by the product of a discrete variable, s_n,
% and a continuous-valued variable, theta_n(t), that is, x_n(t) =
% s_n*theta_n(t).  s_n either takes the value 0 or 1, and follows a 
% Bernoulli distribution with Pr{s_n = 1} = lambda(n).  theta_n(t) follows 
% a complex Gaussian distribution with mean zeta(n) and variance sigma2(n).
% For timesteps t > 1, theta_n(t) evolves according to a Gauss-Markov 
% process, i.e., theta_n(t) = (1 - alpha)*theta_n(t-1) + alpha*w_n(t), 
% where alpha is a scalar between 0 and 1, and w_n(t) is complex Gaussian 
% driving noise with zero mean and variance chosen to keep the steady-state
% variance of the process fixed at kappa_0(n).  Note that the above
% Gauss-Markov evolution expression assumes zeta(n) = 0, and is modified
% for non-zero zeta(n)'s.
%
% Inference is performed using belief propagation (BP) and approximate 
% message passing (AMP).  At each iteration, messages are passed IN
% PARALLEL into each frame, and a recovery of x(t) is produced for each
% frame.  Then, messages are passed out of the frame, and a
% forward/backward pass is conducted to update beliefs on the values of
% {s_n} and {theta(t)}.  This message update schedule is not recommended
% when one wishes to perform (causal) filtering.
%
% SYNTAX:
% [x_hat, v_hat, lambda_hat, StateOut] = sp_mmv_fxn(y, A, Params, ...
%                                                   RunOptions)
%
% INPUTS:
% y                 A 1-by-T cell array of observation vectors.  Each
%                   complex-valued observation vector is defined to be of
%                   length M(t), t = 1, ..., T
% A                 A 1-by-T cell array of complex-valued measurement
%                   matrices, with each matrix of dimension M(t)-by-N, or a
%                   1-by-T cell array of function handles which implement
%                   matrix-vector multiplications.  See below for the form 
%                   of these function handles
% Params            An object of the ModelParams class (see ModelParams.m
%                   in the folder ClassDefs) which contains the parameters
%                   needed to specify the signal/measurement model
% RunOptions        An [optional] object of the Options class (see
%                   Options.m in the ClassDefs folder) that contains
%                   various runtime configurations of AMP-MMV.  If not
%                   provided, a default configuration is used.
%
% OUTPUTS:
% x_hat             A 1-by-T cell array of length-N MMSE estimates of the
%                   unknown time series
% v_hat             A 1-by-T cell array of length-N estimates of the
%                   variances of the coefficient estimates in x_hat
% lambda_hat        A length-N vector estimate of the probabilities of 
%                   activity for s
% StateOut          An object of the AMPState class (see AMPState.m in the
%                   folder ClassDefs) which contains the final state of
%                   AMP-MMV messages.  Useful for warm-starting AMP-MMV on
%                   the same problem but with a different algorithm
%                   configuration.
%
%
% *** If input A is a cell array of function handles, each handle should be
%   of the form @(x,mode). If mode == 1, then the function handle returns 
%   an M(t)-by-1 vector that is the result of A(t)*x.  If mode == 2, then 
%   the function handle returns an N-by-1 vector that is the result of 
%   A(t)'*x. ***
%

%
% Coded by: Justin Ziniel, The Ohio State Univ.
% E-mail: zinielj@ece.osu.edu
% Last change: 12/29/11
% Change summary: 
%		- Created from sp_mmv_fxn v1.0 (12/29/11; JAZ)
% Version 1.1
%

function [x_hat, v_hat, lambda_hat, StateOut] = sp_mmv_fxn(y, A, ...
                                                    Params, RunOptions)

%% Begin by creating a default test case matched to the signal model

if nargin == 0      % No user-provided variables, create default test
    
    % *********************************************************************
    % First generate a signal, measurements, and matrices by using the
    % default configuration specified in SigGenParams.m
    SignalParams = SigGenParams();   % Create default param object
    SignalParams.print();   % Print configuration to command window
    
    % Pass SigGenParams object to the signal generation function
    [x_true, y, A, support, K, sig2e] = signal_gen_fxn(SignalParams);
    
    % Plot true signal
    figure(1); clf; imagesc(abs([x_true{:}])); xlabel('Timestep (t)');
    ylabel('|x_n^{(t)}|'); title('True signal magnitude trajectories');
    colorbar;
    % *********************************************************************
    
    % *********************************************************************
    % Now use the exact same default model parameters to create the
    % required Params input to this function, (see ModelParams.m)
    Params = ModelParams(SignalParams, sig2e);  % Use secondary constructor
    Params.print();     % Print model parameters to command window
    % *********************************************************************

    % ***********************************************************
    % Compute the support-aware Kalman smoother estimate
    disp('Warning: Using AMP instead of BP for support-aware smoother');
    [x_genie, v_genie] = genie_multi_frame_fxn(y, A, support, ...
        Params, Options('smooth_iters', 10, 'alg', 'AMP'));
    for t = 1:SignalParams.T, NMSE_genie(t) = ...
            (norm(x_true{t} - x_genie{t}) / norm(x_true{t}))^2; end
    % Support-aware smoothed genie avg NMSE (dB)
    NMSEavg_dB_genie = 10*log10(sum(NMSE_genie)/SignalParams.T);
    % ***********************************************************
end


%% If user is providing variables, do some basic error checking

if nargin < 3 && nargin > 0
    error('sp_mmv_fxn: Insufficient number of input arguments')
elseif nargin >= 3 || nargin == 0      % Correct number of input arguments
    % Verify dimensionality agreement between y(t) and A(t)
    T = length(y);      % Total # of timesteps
    if isa(A{1}, 'function_handle')
        explicitA = 0;      % Matrix A is not given explicitly
        try
            N = length(A{1}(y{1},2));
        catch
            error('sp_mmv_fxn: Improperly designed function handle')
        end
        for t = 1:T
            M(t) = length(y{t});
        end
    else    % Matrix is provided explicitly
        for t = 1:T
            M(t) = length(y{t});
        end
        explicitA = 1;
        N = size(A{1}, 2);
    end

    % Declare/initialize certain variables
    tol = 1e-1;         % Tolerance for early termination
    filter = 0;         % Not running in filtering mode
    last_iter = 0;      % Flag for early termination
    
    if nargin == 4  % Options object passed
        % Grab the runtime options from the RunOptions input
        if strcmp('Options', class(RunOptions))
            [smooth_iters, min_iters, inner_iters, alg, update, ...
                upd_groups, verbose, StateIn, eps, tau] = ...
                RunOptions.getOptions();
        else
            error('Input to RunOptions must be object of Options class')
        end
    else
        % No Options object passed to function, thus use default options
        Opt = Options();
        Opt.verbose = true;
        Opt.print();    % Print runtime options to command window
        [smooth_iters, min_iters, inner_iters, alg, update, ...
                upd_groups, verbose, StateIn, eps, tau] = Opt.getOptions();
    end

    % Do a little conditioning on the inputs from the RunOptions
    if smooth_iters == -1, smooth_iters = 1; filter = 1; end
    if isempty(upd_groups)
        % No update groups specified, thus assign all coefficients to the
        % same update group (for EM param learning)
        upd_groups = {1:N};
    else    % User has specified custom update groups.  Check validity
        if ~numel(setdiff(1:N, cat(1,upd_groups{:}))) == 0
            error(['''RunOptions.upd_groups'' does not partition every' ...
                ' coefficient into a unique group'])
        end
    end

    % Unpack 'Params' structure
    if strcmp('ModelParams', class(Params))
        [lambda, zeta, sigma2, alpha, sig2e, rho] = Params.getParams();
    else
        error('Input to Params must be object of ModelParams class');
    end
    
    % Check to make sure dimensions of the parameters are okay, vectorize
    % if necessary
    if numel(lambda) == 1, lambda = lambda*ones(N,1);
    elseif size(lambda,1) ~= N, error('Incorrect dimension: lambda'); end
    if numel(zeta) == 1, zeta = zeta*ones(N,1);
    elseif size(zeta,1) ~= N, error('Incorrect dimension: zeta'); end
    if numel(sigma2) == 1, sigma2 = sigma2*ones(N,1);
    elseif size(sigma2,1) ~= N, error('Incorrect dimension: sigma2'); end
    if numel(alpha) == 1, alpha = alpha*ones(N,1);
    elseif size(alpha,1) ~= N, error('Incorrect dimension: alpha'); end
    if numel(rho) == 1, rho = rho*ones(N,1);
    elseif size(rho,1) ~= N, error('Incorrect dimension: rho'); end
    if numel(sig2e) ~= 1, error('Incorrect dimension: sig2e'); end
    
end  % End of input error checking/unpacking


%% Run the AMP-MMV algorithm w/ the parallel message passing schedule

% ********************************************************************************************
% Would be nice to use user-specified parameters to initialize parameters
% at each timestep

% Message matrix declarations and initializations (all are N-by-T dim)
ETA_FWD = [zeta*ones(1,T)];     	% Matrix of means from d(t) to theta(t)
ETA_BWD = zeros(N,T);            	% Matrix of means from d(t+1) to theta(t)
KAPPA_FWD = [sigma2*ones(1,T)];     % Matrix of vars from d(t) to theta(t)
KAPPA_BWD = inf*ones(N,T);        	% Matrix of vars from d(t+1) to theta(t)
PI_IN = lambda*ones(1,T);          	% Matrix of messages from s to f(t)
PI_OUT = NaN*ones(N,T);         	% Matrix of messages from f(t) to s
XI_IN = NaN*ones(N,T);            	% Matrix of means from theta(t) to f(t)
XI_OUT = NaN*ones(N,T);          	% Matrix of means from f(t) to theta(t)
PSI_IN = NaN*ones(N,T);           	% Matrix of vars from theta(t) to f(t)
PSI_OUT = NaN*ones(N,T);           	% Matrix of vars from f(t) to theta(t)

% Create space to store the state of the AMP variables between iterations
if ~isempty(StateIn)    % Warm-start AMP-MMV
    [C_STATE, Z_State, MU_STATE, ETA_FWD, ETA_BWD, KAPPA_FWD, ...
        KAPPA_BWD, PI_IN, PI_OUT] = StateIn.getState();
else
    C_STATE = 100*mean(sigma2)*ones(1,T);	% Initialize to large c values
    Z_State = y;                % Initialize residual to the measurements
    MU_STATE = zeros(N,T);
end

% Make room for some variables that will be filled in later
x_hat = cell(1,T);                   	% Placeholder for MMSE estimates
x_old = inf*ones(N,T);                 	% Will hold previous estimates
V_STATE = NaN*ones(N,T);             	% Placeholder for x variances
NMSEavg_dB = NaN*ones(1,smooth_iters); 	% Placeholder for alg. avg. NMSE (dB)

% Declare constants
junk = NaN*ones(N,1);       % Pass this variable when input is irrelevant
Eq_Params.eps = eps;        % Gaussian "squelch" parameter on p(x_n)
Eq_Params.tau = tau;        % Type of approx to use for f-to-theta msgs
Eq_Params.sig2e = sig2e;    % Circular, complex, additive, Gaussian noise variance
Eq_Options.iter = inner_iters;  % Number of BP/AMP iterations per equalization run
Eq_Options.alg = alg;       % Specifies BP- or AMP-driven inference
Support.pz1 = 0;            % Irrelevant parameter for MMV problem
Support.p1z = 0;            % Irrelevant parameter for MMV problem
Amplitude.alpha = alpha;    % Amplitude Gauss-Markov process innovation rate
Amplitude.rho = rho;        % Amplitude Gauss-Markov process driving noise

start_time = tic;       % Start stopwatch

% Execute the message passing routine
for k = 1:smooth_iters       % Iteration of the forwards/backwards message pass
    
    % Begin by computing the messages moving into each frame.  Then obtain
    % recoveries of x(t) for every timestep "in parallel"
    for t = 1:T           % Current timestep index
        % At the current timestep, calculate messages going from s and
        % theta(t) to f(t)
        if t < T      % Not the terminal timestep, so msgs are multiplied
            [~, XI_IN(:,t), PSI_IN(:,t)] = ...
                sp_msg_mult_fxn(junk, junk, ETA_FWD(:,t), ...
                KAPPA_FWD(:,t), ETA_BWD(:,t), KAPPA_BWD(:,t));
        else            % Terminal timestep, thus just pass the lone quantities
            XI_IN(:,t) = ETA_FWD(:,t);
            PSI_IN(:,t) = KAPPA_FWD(:,t);
        end
        
        % Perform equalization to obtain an estimate of x(t) using the
        % priors specified by the current values of PI_IN(:,t), XI_IN(:,t),
        % and PSI_IN(:,t).  Update the outoing quantities, and save current
        % best estimate of x(t)
        Eq_Params.pi = PI_IN(:,t); Eq_Params.xi = XI_IN(:,t); 
        Eq_Params.psi = PSI_IN(:,t);
        EqStateIn.c = C_STATE(t); EqStateIn.z = Z_State{t};
        EqStateIn.mu = MU_STATE(:,t);
        % ****************************************************************
        [x_hat{t}, OutMessages, StateOutAMP] = sp_frame_fxn(y{t}, A{t}, ...
            Eq_Params, Eq_Options, EqStateIn);
        % ****************************************************************
        PI_OUT(:,t) = OutMessages.pi; XI_OUT(:,t) = OutMessages.xi;
        PSI_OUT(:,t) = OutMessages.psi;     % Updated message parameters
        C_STATE(t) = StateOutAMP.c; Z_State{t} = StateOutAMP.z;
        MU_STATE(:,t) = StateOutAMP.mu;    % Save state of AMP for warm start
        V_STATE(:,t) = StateOutAMP.v;
    end
    
    % Now use the outgoing messages from each frame to conduct a
    % forward/backward pass to update beliefs about the values of each s
    % and theta(t).  Start with the forward portion of the pass
    for t = 1:T       % Current timestep index (ahead by 1)
        % Use the outgoing messages from f(t) to s and
        % theta(t) to update the priors on s and theta(t+1)
        if t < T      % We aren't at the terminal timestep, so update fwd quantities
            Support.lambda = junk;              % Msg. irrelevant
            Support.pi = junk;                  % Msg. irrelevant
            Amplitude.eta = ETA_FWD(:,t);       % Msg. d(t) to theta(t)
            Amplitude.kappa = KAPPA_FWD(:,t);   % Msg. d(t) to theta(t)
            Amplitude.xi = XI_OUT(:,t);         % Msg. f(t) to theta(t)
            Amplitude.psi = PSI_OUT(:,t);       % Msg. f(t) to theta(t)
            Amplitude.eta_0 = ETA_FWD(:,1);     % Mean of the amplitude RP
            Msg.direction = 'forward';          % Indicate forward propagation
            Msg.terminal = 0;                   % Non-terminal update
            
            % Compute the updates
            [~, ETA_FWD(:,t+1), KAPPA_FWD(:,t+1)] = ...
                sp_timestep_fxn(Support, Amplitude, Msg);
            
            % Update all the PI_INs for the next round
            inc_ind = [1:t-1, t+1:T];     % Timesteps in the PI_OUT prod.
            PI_IN(:,t) = lambda.*prod(PI_OUT(:,inc_ind), 2) ./ ...
                (lambda.*prod(PI_OUT(:,inc_ind), 2) + ...
                (1 - lambda).*prod(1 - PI_OUT(:,inc_ind), 2));
        else            % Terminal timestep, thus update backwards priors
            Support.lambda = junk;              % Msg. irrelevant
            Support.pi = junk;                  % Msg. irrelevant
            Amplitude.eta = NaN*ones(N,1);    	% Msg. D.N.E.
            Amplitude.kappa = NaN*ones(N,1);    % Msg. D.N.E.
            Amplitude.xi = XI_OUT(:,t);         % Msg. f(t) to theta(t)
            Amplitude.psi = PSI_OUT(:,t);       % Msg. f(t) to theta(t)
            Amplitude.eta_0 = ETA_FWD(:,1);     % Mean of the amplitude RP
            Msg.direction = 'backward';      	% Indicate forward propagation
            Msg.terminal = 1;                   % Terminal update            
            
            % Compute the updates
            [~, ETA_BWD(:,t-1), KAPPA_BWD(:,t-1)] = ...
                sp_timestep_fxn(Support, Amplitude, Msg);
            
            % Update PI_IN(:,T) for the next round
            inc_ind = 1:T-1;
            PI_IN(:,t) = lambda.*prod(PI_OUT(:,inc_ind), 2) ./ ...
                (lambda.*prod(PI_OUT(:,inc_ind), 2) + ...
                (1 - lambda).*prod(1 - PI_OUT(:,inc_ind), 2));
        end
        
        % On final pass, save the estimates of the coefficient variances
        if (k == smooth_iters) || last_iter
            if t == T
                v_hat{t} = StateOutAMP.v;              % var{x_n(t) | y(t)}
                lambda_hat = lambda.*prod(PI_OUT, 2) ./ ...
                    (prod((1 - PI_OUT), 2).*(1 - lambda) ...
                    + prod(PI_OUT, 2).*lambda);
                if filter, return; end; % Early break for filter mode
            elseif filter && (t < T)
                % Finalize quantities for filter mode output
                v_hat{t} = StateOutAMP.v;      % var{x_n(t) | y(t)}
            end
        end
    end
    
    % Now execute the backwards portion of the message passing for all the 
    % s and theta(t) variables (t = 0, ..., T-1)
    for t = T-1:-1:1           % Descend from 2nd-to-last timestep to the first
        % Use the outgoing messages from f(t) to s(t) and
        % theta(t) to update the priors on s(t+1) and theta(t+1)
        if t > 1        % We aren't at the first timestep, so update bwd quantities
            Support.lambda = junk;              % Msg. irrelevant
            Support.pi = junk;                  % Msg. irrelevant
            Amplitude.eta = ETA_BWD(:,t);       % Msg. d(t) to theta(t)
            Amplitude.kappa = KAPPA_BWD(:,t);   % Msg. d(t) to theta(t)
            Amplitude.xi = XI_OUT(:,t);         % Msg. f(t) to theta(t)
            Amplitude.psi = PSI_OUT(:,t);       % Msg. f(t) to theta(t)
            Amplitude.eta_0 = ETA_FWD(:,1);     % Mean of the amplitude RP
            Msg.direction = 'backward';      	% Indicate backward propagation
            Msg.terminal = 0;                   % Non-terminal update
            
            % Compute the updates
            [~, ETA_BWD(:,t-1), KAPPA_BWD(:,t-1)] = ...
                sp_timestep_fxn(Support, Amplitude, Msg);
        else            % Initial timestep, thus there is nothing to update
            % Nothing to do in here since eta_0, and kappa_0 do
            % not change
        end
        
        % On final pass, save the estimates of the coefficient variances
        if k == smooth_iters || last_iter
            v_hat{t} = StateOutAMP.v;              % var{x_n(t) | {y(t)}}
        end
    end

    
    % Compute the time-averaged residual energy
    avg_resid_energy = 0;
    for t = 1:T
        if explicitA
            avg_resid_energy = avg_resid_energy + ...
                norm(y{t} - A{t}*x_hat{t})^2 / T;
        else
            avg_resid_energy = avg_resid_energy + ...
                norm(y{t} - A{t}(x_hat{t}, 1))^2 / T;
        end
    end
    
    % Display execution info, if requested
    if verbose
        fprintf('SP_MMV_FXN: Completed %d iterations\n', k);
        fprintf('Total elapsed time: %f s\n', toc(start_time));
        fprintf('Time-averaged residual energy: %f\n', avg_resid_energy);
        fprintf('---------------------------------------------------------\n');
    end
    
    % ********************************************************************
    % Having completed a forward\backward pass, update the hyperparameters
    % using the current recovery, if the user wishes.  Begin only after two
    % smoothing iterations have elapsed, to avoid making inferences based
    % on poor initial signal estimates
    % ********************************************************************
    if update && k > 2
        % Pass current state of factor graph to the parameter update
        % function through this structure
        State = struct('Mu_x', []);     % Clear contents of structure
        State.Mu_x = MU_STATE;      State.V_x = V_STATE;
        State.Eta_fwd = ETA_FWD;    State.Eta_bwd = ETA_BWD;
        State.Xi_out = XI_OUT;      State.Kap_fwd = KAPPA_FWD;
        State.Kap_bwd = KAPPA_BWD;  State.Psi_out = PSI_OUT;
        State.Pi_out = PI_OUT;      State.iter = k;
        
        % Also pass the current values of the parameters
        UpdParams = struct('eta_0', []);    % Clear contents of structure
        UpdParams.eta_0 = ETA_FWD(:,1);     
        UpdParams.alpha = Amplitude.alpha;
        UpdParams.rho = Amplitude.rho;
        UpdParams.lambda = lambda;
        UpdParams.sig2e = Eq_Params.sig2e;
        
        %Now compute the updates
        Updates = mmv_parameter_update_fxn(y, A, State, UpdParams, upd_groups);
        
        % Store the updated hyperparameters in needed locations
        for g = 1:length(upd_groups)
            lambda(upd_groups{g},1) = Updates.lambda(g);
            KAPPA_FWD(upd_groups{g},1) = Updates.alpha(g)*Updates.rho(g) / ...
                (2 - Updates.alpha(g));
            ETA_FWD(upd_groups{g},1) = Updates.eta_0(g);
            Amplitude.alpha(upd_groups{g}) = .01*Amplitude.alpha(upd_groups{g}) + ...
                .99*Updates.alpha(g)*ones(length(upd_groups{g}),1);
            Amplitude.rho(upd_groups{g},1) = Updates.rho(g);
        end
        Eq_Params.sig2e = .01*Eq_Params.sig2e + .99*Updates.sig2e; 
        
        if verbose
            for g = 1:length(upd_groups)
                disp(sprintf('Updated value of ''lambda'' for Group %d: %f', ...
                    g, Updates.lambda(g)));
                disp(sprintf('Updated value of ''alpha'' for Group %d: %f', ...
                    g, Amplitude.alpha(upd_groups{g}(1))));
                disp([sprintf('Updated value of ''zeta'' for Group %d: ', ...
                    g) num2str(Updates.eta_0(g))]);
                disp(sprintf('Updated value of ''sigma2'' for Group %d: %f', ...
                    g, KAPPA_FWD(upd_groups{g}(1),1)));
            end
            disp(sprintf('Updated value of ''sig2e'': %f', Eq_Params.sig2e));
            disp('---------------------------------------------------------');
        end
    end
    
   	% If this is a default test case, then plot the result
    if nargin == 0
        for l = 1,      % Plot various things
            % First plot the current recovery
            figure(2); imagesc(abs([x_hat{:}])); colorbar; 
            title(['MMSE estimate | Fwd./Bwd. iters: ' num2str(k) '/' num2str(k)]);
            xlabel('Timestep (t)'); ylabel('|x_{mmse}(t)|');
            K_handle = line(1:T, K); 
            set(K_handle, 'Color', 'Cyan', 'LineStyle', '-.'); 
            M_handle = line(1:T, M);
            set(M_handle, 'Color', 'White', 'LineStyle', '--'); 
            legend_handle = legend('K(t)', 'M(t)');
            set(legend_handle, 'Color', [.392 .475 .635])
            
            % Next plot the NMSEs of the different recovery methods
            for t = 1:T; NMSE(t) = (norm(x_true{t} - ...
                    x_hat{t})/norm(x_true{t}))^2; end
            NMSEavg_dB(k) = 10*log10(sum(NMSE)/T);    % Alg. avg NMSE (dB)
            figure(3);
            plot([1 : 1 : smooth_iters], NMSEavg_dB); hold on
            genie_line = line([0, smooth_iters], [NMSEavg_dB_genie, NMSEavg_dB_genie]);
            set(genie_line, 'Color', 'Green');  hold off
            legend('AMP-MMV NMSE', 'SKS NMSE')
            xlabel('Iteration'); ylabel('Avg. NMSE [dB]');
            title(['Avg. NMSEs | Iterations: ' num2str(k)]);
        end
    end
    
    % Check for early termination this round
    if last_iter, break; end
    
    % Check for early termination for next round
    if k > min_iters && ...
            norm([x_hat{:}] - x_old, 'fro')/norm([x_hat{:}], 'fro') < tol && ...
            avg_resid_energy < sum(M)*sig2e/T
        last_iter = 1;      % Set the flag for last iteration
    else
        x_old = [x_hat{:}];
    end

    % Report to user if maximum number of smoothing iterations was exceeded
    if verbose && k == smooth_iters
        fprintf('sp_mmv_fxn: Max # of smoothing iterations reached\n')
    end
end

% Save AMP-MMV final state as an AMPState object (see AMPState.m)
StateOut = AMPState(C_STATE, Z_State, MU_STATE, ETA_FWD, ETA_BWD, ...
    KAPPA_FWD, KAPPA_BWD, PI_IN, PI_OUT);