% SP_MMV_WRAPPER_FXN    This function acts as a wrapper for calling the
% main AMP-MMV recovery functions, SP_MMV_FXN and SP_MMV_SERIAL_FXN.
% Since, for finite-dimensional problems, the message passing procedure
% is not guaranteed to converge, it can be beneficial to try alternative
% message schedules and/or parameter configurations when an initial choice
% does not appear to be working.  This wrapper function will first attempt
% the parallel message schedule (SP_MMV_FXN) using a Taylor series
% approximation of the outgoing AMP amplitude messages, (i.e., Params.tau =
% -1 in SP_MMV_FXN), and the parameters provided by the user.  If the
% residual energy of the resultant recovery greatly exceeds the expected
% residual energy, given the noise variance, then it is assumed that the
% parallel message schedule failed to converge, in which case the serial
% message schedule will be applied (SP_MMV_SERIAL_FXN).  In this manner,
% schedules will be alternated, and iteration limits will be adjusted,
% until a recovery is obtained that has a small enough residual energy, or
% a maximum number of attempts has been reached.
%
% NOTE: For simplicity, the various schedule configurations have been
% hard-coded into this function, rather than passed as additional
% arguments.  If you wish to modify these configurations, please scroll
% down, where you will find a clearly designated section that can be
% modified to suit your particular needs.
%
% SYNTAX:
% [x_hat, v_hat, lambda_hat] = sp_mmv_wrapper_fxn(y, A, Params, RunOptions)
%
% INPUTS:
% y                 A 1-by-T cell array of observation vectors.  Each
%                   complex-valued observation vector is defined to be of
%                   length M(t), t = 1, ..., T
% A                 A 1-by-T cell array of complex-valued measurement
%                   matrices, with each matrix of dimension M(t)-by-N, or a
%                   1-by-T cell array of function handles which implement
%                   matrix-vector multiplications.  See below for the form 
%                   of these function handles
% Params            An object of the ModelParams class (see ModelParams.m
%                   in the folder ClassDefs) which contains the parameters
%                   needed to specify the signal/measurement model
% RunOptions        An [optional] object of the Options class (see
%                   Options.m in the folder ClassDefs) that contains
%                   various runtime configurations of AMP-MMV.  If not
%                   provided, a default configuration is used.
%
% OUTPUTS:
% x_hat             A 1-by-T cell array of length-N MMSE estimates of the
%                   unknown time series
% v_hat             A 1-by-T cell array of length-N estimates of the
%                   variances of the coefficient estimates in x_hat
% lambda_hat        A length-N vector estimate of the probabilities of 
%                   activity for s
%
%
% *** If input A is a cell array of function handles, each handle should be
%   of the form @(x,mode). If mode == 1, then the function handle returns 
%   an M(t)-by-1 vector that is the result of A(t)*x.  If mode == 2, then 
%   the function handle returns an N-by-1 vector that is the result of 
%   A(t)'*x. ***
%

%
% Coded by: Justin Ziniel, The Ohio State Univ.
% E-mail: zinielj@ece.osu.edu
% Last change: 12/29/11
% Change summary: 
%		- Created from sp_mmv_wrapper_fxn v1.0 (12/29/11; JAZ)
% Version 1.1
%

function [x_hat, v_hat, lambda_hat] = sp_mmv_wrapper_fxn(y, A, Params, ...
                                                         RunOptions)

%% Create a default test case matched to the signal model if no user input

if nargin == 0      % No user-provided variables, create default test
    
    % *********************************************************************
    % First generate a signal, measurements, and matrices by using the
    % default configuration specified in SigGenParams.m
    SignalParams = SigGenParams();   % Create default param object
    SignalParams.N = 5000;
    SignalParams.M = 1563;
    SignalParams.lambda = 0.15;
    SignalParams.print();   % Print configuration to command window
    
    % Pass SigGenParams object to the signal generation function
    [x_true, y, A, support, K, sig2e] = signal_gen_fxn(SignalParams);
    
    % Plot true signal
    figure(1); clf; imagesc(abs([x_true{:}])); xlabel('Timestep (t)');
    ylabel('|x_n^{(t)}|'); title('True signal magnitude trajectories');
    colorbar;
    % *********************************************************************
    
    % *********************************************************************
    % Now use the exact same default model parameters to create the
    % required Params input to this function, (see ModelParams.m)
    Params = ModelParams(SignalParams, sig2e);  % Use secondary constructor
    Params.print();     % Print model parameters to command window
    % *********************************************************************

%     % ***********************************************************
%     % Compute the support-aware smoothed genie estimate
%     disp('Warning: Using AMP instead of ABP for support-aware smoother');
%     [x_genie, v_genie] = genie_multi_frame_fxn(y, A, support, ...
%         Params, struct('smooth_iter', 10, 'inner_iters', ...
%         inner_iters, 'alg', 2));
%     % ***********************************************************

end


%% If user is providing variables, do some basic error checking

if nargin < 3 && nargin > 0
    error('sp_mmv_wrapper_fxn: Insufficient number of input arguments')
elseif nargin >= 3 || nargin == 0	% Correct number of input arguments
    
    % Verify dimensionality agreement between y(t) and A(t)
    T = length(y);      % Total # of timesteps
    if isa(A{1}, 'function_handle')
        explicitA = 0;      % Matrix A is not given explicitly
        try
            N = length(A{1}(y{1},2));
        catch
            error('sp_mmv_serial_fxn: Improperly designed function handle')
        end
        for t = 1:T
            M(t) = length(y{t});
        end
    else    % Matrix is provided explicitly
        for t = 1:T
            M(t) = length(y{t});
        end
        explicitA = 1;
        N = size(A{1}, 2);
    end
    
    if nargin == 4  % Options object passed
        % Grab the verbosity from the input RunOptions
        if strcmp('Options', class(RunOptions))
            wrap_verbose = RunOptions.verbose;
        else
            error('Input to RunOptions must be object of Options class')
        end
    else
        % No Options object passed to function, thus use default options
        RunOptions = Options();
        RunOptions.verbose = true;
        RunOptions.print();    % Print runtime options to command window
        wrap_verbose = RunOptions.verbose;
    end
    
    
    % *********************************************************************
    % *********************************************************************
    % ******    Define the various schedule configurations below     ******
    % *********************************************************************
    % *********************************************************************
    schedule = [1 2 1 2 1 2 1 2];   % Parallel: (1), Serial: (2)
    smooth_iters = [RunOptions.smooth_iters*ones(1,4), 50, 50, 100, 100];
    min_iters = [RunOptions.min_iters*ones(1,4), 5, 5, 5, 5];
    inner_iters = [RunOptions.inner_iters*ones(1,4), 5, 5, 5, 5];
    tau = [RunOptions.tau*ones(1,4), -ones(1,4)];
    % *********************************************************************
    % *********************************************************************
    % ******    Define the various schedule configurations above     ******
    % *********************************************************************
    % *********************************************************************
    
    
    % Residual energy multiplier that triggers the use of alternative
    % schedule(s) and configuration(s)
    resid_mult = 1.25;
    
end  % End of input error checking/unpacking


%% Recover the signal

% Start by defining the possible different runtime configurations that will
% be cycled through (if residual energy exceeds the multiplier)
Ncycles = numel(schedule);
cyc = 1;
if Ncycles ~= numel(smooth_iters) || Ncycles ~= numel(min_iters) || ...
        Ncycles ~= numel(inner_iters) || Ncycles ~= numel(tau)
    error('Please verify dimensionality agreement of configurations')
end
schedtext = {'parallel', 'serial'};     % Text labels

tstart = tic;   % Start stopwatch

% Create storage space for the output variables.  Ultimately, only one cell
% of each array will be returned
x_hat = cell(1,Ncycles);
v_hat = cell(1,Ncycles);
lambda_hat = cell(1,Ncycles);

TAresid = inf*ones(1,Ncycles);  % Time-averaged residual energy

if wrap_verbose
    fprintf(['************************************************' ...
        '*************************************************\n']);
end

% Start cycling through the various runtime configurations until one is
% found that has a time-averaged residual energy that falls below the
% target residual energy level
while min(TAresid) > resid_mult*sum(M)*Params.sig2e/T && cyc <= Ncycles
    
    % For this configuration, modify the RunOptions structure that will be
    % passed to the recovery algorithm
    RunOptions.smooth_iters = smooth_iters(cyc);
    RunOptions.min_iters = min_iters(cyc);
    RunOptions.inner_iters = inner_iters(cyc);
    RunOptions.tau = tau(cyc);
    
    % Now run the appropriate message schedule
    try
        if wrap_verbose
            if tau(cyc) == -1
                fprintf(['Running %s schedule w/ Taylor approx. & ' ...
                    'smooth_iters = %d, min_iters = %d, inner_iters = %d\n'], ...
                    schedtext{schedule(cyc)}, smooth_iters(cyc), ...
                    min_iters(cyc), inner_iters(cyc));
            else
                fprintf(['Running %s schedule w/ tau = 1-%1.1e, ' ...
                    'smooth_iters = %d, min_iters = %d, inner_iters = %d\n'], ...
                    schedtext{schedule(cyc)}, 1-tau(cyc), smooth_iters(cyc), ...
                    min_iters(cyc), inner_iters(cyc));
            end
        end
        
        
        % Here's where we actually call AMP-MMV according to one of the two
        % schedules
        % *****************************************************************
        if schedule(cyc) == 1   % Parallel schedule

            [x_hat{cyc}, v_hat{cyc}, lambda_hat{cyc}, RunOptions.StateIn] = ...
                sp_mmv_fxn(y, A, Params, RunOptions);

        elseif schedule(cyc) == 2   % Serial schedule

            [x_hat{cyc}, v_hat{cyc}, lambda_hat{cyc}, RunOptions.StateIn] = ...
                sp_mmv_serial_fxn(y, A, Params, RunOptions);
        
        else
            error('Invalid assignment: schedule (should be 1 or 2)')
        end
        % *****************************************************************
        
        % Clear the message states if changing number of smoothing iters
        if cyc < Ncycles
            if smooth_iters(cyc) ~= smooth_iters(cyc+1)
                RunOptions.StateIn = [];
            end
        end
        
        % Compute the time-averaged residual energy
        avg_resid_energy = 0;
        for t = 1:T
            if explicitA
                avg_resid_energy = avg_resid_energy + ...
                    norm(y{t} - A{t}*x_hat{cyc}{t})^2 / T;
            else
                avg_resid_energy = avg_resid_energy + ...
                    norm(y{t} - A{t}(x_hat{cyc}{t}, 1))^2 / T;
            end
        end
        TAresid(cyc) = avg_resid_energy;
        
        if wrap_verbose
            fprintf('Residual energy ratio (actual/expected): %g\n', ...
                TAresid(cyc) / (sum(M)*Params.sig2e/T));
        end
    catch Exception
        % Something went wrong with this configuration.  We'll try another,
        % and hope for the best...
        fprintf([Exception.message '\n']);
    end
    
    % Increment cycle counter
    cyc = cyc + 1;
end

% Now we have either run just enough schedules to find a recovery with a 
% low enough residual energy, or else we went through all the possible
% configurations without finding such a recovery.  Either way, return the
% recovery with the lowest residual energy
[TAresid, mincyc] = min(TAresid);
x_hat = x_hat{mincyc};
v_hat = v_hat{mincyc};
lambda_hat = lambda_hat{mincyc};

if wrap_verbose
    if tau(mincyc) == -1
        fprintf(['Returning recovery from %s schedule w/ Taylor approx.' ...
            ' & smooth_iters = %d, min_iters = %d, inner_iters = %d\n'], ...
            schedtext{schedule(mincyc)}, smooth_iters(mincyc), ...
            min_iters(mincyc), inner_iters(mincyc));
    else
        fprintf(['Returning recovery from %s schedule w/ tau = 1-%e, ' ...
            'smooth_iters = %d, min_iters = %d, inner_iters = %d\n'], ...
            schedtext{schedule(mincyc)}, 1-tau(mincyc), smooth_iters(mincyc), ...
            min_iters(mincyc), inner_iters(mincyc));
    end
    fprintf(['************************************************' ...
    '*************************************************\n']);
end

if nargin == 0  % Display some performance metrics
    TNMSE = sum(sum(([x_true{:}] - [x_hat{:}]).^2, 1) ./ ...
        sum([x_true{:}].^2, 1))/T;
    fprintf(['TNMSE = %2.2g dB  |  Trial Residual Energy = %g' ...
        '  |  Trial time: %g\n'], 10*log10(TNMSE), TAresid, ...
        toc(tstart));
end