% CHANGING_MTX_TEST 	A test of MPI-MMV and the support-aware Kalman
% smoother (SKS) that will study the performance benefits that are obtained
% by allowing the MMV measurement matrix to change over time.  In this
% test, the measurement matrix is assumed to have entries that are i.i.d.
% complex Gaussian distributed.  At each timestep, the measurement matrix
% evolves according to a first-order Gauss-Markov random process dynamic,
% namely
%                 A(t) = (1 - beta) A(t-1) + beta U(t),
% where beta is a scalar that controls how quickly A(t) changes over time,
% and U(t) is a complex Gaussian matrix that is independent of A(t-1), with
% a variance that is chosen to preserve (in expectation) unit-norm columns 
% of A(t).
% For this test, beta will be swept over a range of values in the interval
% (0,1], and NSER, TNMSE, and runtime information will be captured.
%
%
% SYNTAX:
% changing_mtx_test(N, M, T, lambda, SNRdB, N_trials)
%
% INPUTS:
%   N               Number of unknowns at each timestep
%   M               Number of measurements at each timestep
%   T               Number of timesteps
%   lambda          Activity probability (Pr{s_n = 1})
%   SNRdB           SNR at each timestep (in dB)
%   N_trials     	Number of trials to average results over
%
% OUTPUTS:
% This function will save the timestep-averaged normalized mean-squared
% error (TNMSE) and the normalized support error rate (NSER), measuring the
% similarity between the estimated and true support set.  Results are 
% stored in a .MAT file whose filename is created at runtime based on input
% parameters and the time/date of the start of execution.
% Suffix keywords:
%   _MPI        The multi-timestep BP algorithm (AMP-MMV)
%   _SKS        The support-aware Kalman smoother
%

%
% Coded by: Justin Ziniel, The Ohio State Univ.
% E-mail: ziniel.1@osu.edu
% Last change: 12/29/11
% Change summary: 
%		- Created from changing_mtx_test.m v1.0 (12/29/11; JAZ)
% Version 1.1
%


function changing_mtx_test(N, M, T, lambda, SNRdB, N_trials)

%% Declare the parameters for the signal model and test setup

if nargin == 0
    N = 1024;            	% # of parameters
    M = round((1/30)*N);  	% # of measurements
    T = 4;                  % # of timesteps
    lambda = (1/2)*(M/N);   % Activity probability (Pr{s_n = 1})
    SNRdB = 25;             % SNR at each timestep (in dB)
    N_trials = 25;       	% Number of trials to average results over
end

warning('SKS is being run using AMP instead of ABP')

% Test setup parameters
alpha_sweep = [0.01];        % Correlation coefficients to use
beta_sweep = logspace(-2, 0, 10);     % Pr{s_n = 1} values
N_alpha = numel(alpha_sweep);
N_beta = numel(beta_sweep);

% Important BP execution parameters
inner_iters = 15; 	% Max # of inner AMP iterations
smooth_iters = 5;   % Max # of smoothing iterations
alg = 'AMP';       	% Use AMP, not BP
update = true;   	% Use EM parameter learning
verbose = true;   	% SP_MMV_WRAPPER_FXN verbosity
eps = 1e-7;         % Regularization parameter

% Filename for saving data
savefile = [datestr(now, 'ddmmmyyyy_HHMMSSFFF') '_changing_mtx_test_N_' ...
    num2str(N) '_M_' num2str(M) '_T_' num2str(T) '_lambda_' ...
    num2str(lambda) '_SNRdB_' num2str(SNRdB) '.mat'];

% Randomly seed the RNG and save its state for reference purposes
NewStream = RandStream.create('mt19937ar', 'Seed', sum(100*clock));
RandStream.setDefaultStream(NewStream);
savedState = NewStream.State;


%% Execute the phase plane test

% Create arrays to store the failure rates and MSEs
NSER_MPI = NaN*ones(N_alpha, N_beta, N_trials);  	% NSER for MPI-MMV
NSER2_MPI = NaN*ones(N_alpha, N_beta, N_trials);  	% 2nd NSER for MPI-MMV
NSER_SKS = NaN*ones(N_alpha, N_beta, N_trials);    	% NSER for SKS
TNMSE_MPI = NaN*ones(N_alpha, N_beta, N_trials);    % TNMSE for MPI-MMV
TNMSE_SKS = NaN*ones(N_alpha, N_beta, N_trials);   	% TNMSE for SKS
Run_tMSBL = NaN*ones(N_alpha, N_beta, N_trials);    % Runtime for T-MSBL
Run_tSBL = NaN*ones(N_alpha, N_beta, N_trials);   	% Runtime for T-SBL
Run_MSBL = NaN*ones(N_alpha, N_beta, N_trials);     % Runtime for MSBL
Run_MPI = NaN*ones(N_alpha, N_beta, N_trials);      % Runtime for MPI-MMV
Run_SKS = NaN*ones(N_alpha, N_beta, N_trials);      % Runtime for SKS

% try % Matlabpool opening
%     % Open a matlabpool, if one is not already open
%     if matlabpool('size') == 0
%         matlabpool('open', min(N_beta, 8));
%         disp(sprintf('Opened a matlabpool with %d workers', matlabpool('size')));
%     end
% catch
%     disp('Error in opening a matlabpool')
% end

time = clock;   % Start the stopwatch
for n = 1:N_trials
    for a = 1:N_alpha     % Chosen alpha
        alpha = alpha_sweep(a);
        for b = 1:N_beta  	% # of active coefficients
            beta = beta_sweep(b);
            
            % *************************************************************
            % Generate a signal/measurement pair realization for this suite
            % according to the MPI-MMV signal model
            % *************************************************************
            % Construct the signal generation parameter object
            GenParams = SigGenParams();  % Start w/ default object
            GenParams.N = N;             GenParams.M = M;
            GenParams.T = T;             GenParams.lambda = lambda;   
            GenParams.zeta = 0;          GenParams.sigma2 = 1;        
            GenParams.alpha = alpha;     GenParams.SNRmdB = SNRdB;    
            GenParams.version = 'C';     GenParams.A_type = 1;

            % ***********************************************************
            % Generate the time series according to the above parameters
            % using the modified signal generation function that will
            % produce the Gauss-Markov measurement matrices
            [x_true, y, A, true_support, K, sig2e] = ...
                signal_gen_fxn_mod(GenParams, beta);
            if isnan(sig2e)
                [x_true, y, A, true_support, K, sig2e] = ...
                    signal_gen_fxn_mod(GenParams, beta);
            end
            
            % Create variables based on the realization that better match
            % the signal model of T-MSBL, T-SBL, and MSBL
            W_TRUE = [x_true{:}];
            Y = [y{:}];
            Phi = A{1};            

            % *************************************************************
            %       Solve using the support-aware Kalman smoother
            % *************************************************************            
            % Create the signal model parameter object
            Params = ModelParams(GenParams, sig2e);

            % Create an object to hold SKS runtime options
            RunOptions = Options();     %Start with default object
            RunOptions.inner_iters = inner_iters;
            RunOptions.smooth_iters = smooth_iters;
            % (Defaults to ABP mode, not AMP mode)
            RunOptions.alg = alg;      RunOptions.eps = eps;
            
            if T == 1   % Need to use sp_frame_fxn only
                Params.pi = zeros(N,1);     Params.pi(true_support) = 1;
                Params.xi = GenParams.zeta;	
                Params.psi = GenParams.sigma2;
                Params.eps = eps;
                EqOptions.iter = inner_iters;  EqOptions.alg = 2;
                
                tic;
                X_sks = cell(1,1);
                % Solve using sp_frame_fxn (single frame)
                X_sks{1} = sp_frame_fxn(Y, Phi, Params, EqOptions);
                Run_SKS(a,b,n) = toc;
            else
                tic;
                % Solve using the SKS
                [X_sks] = genie_multi_frame_fxn(y, A, true_support, Params, RunOptions);
                Run_SKS(a,b,n) = toc;
            end
            
            % Time-averaged normalized MSE (TNMSE)
            TNMSE_SKS(a,b,n) = sum(sum(abs(W_TRUE - [X_sks{:}]).^2, 1)./sum(abs(W_TRUE).^2, 1))/T;
            
            % NSER
            NSER_SKS(a,b,n) = nser(true_support, ...
                find(sum(abs([X_sks{:}]),2) ~= 0));
            


            % *************************************************************
            %                       Solve using AMP-MMV
            % *************************************************************
            % Create a structure to hold the signal model parameters
            Params = ModelParams(GenParams, sig2e);
            
            % Try to guess some decent initial values for the parameters
            if T > 1
                zeta_init = 0;
                sig2e_init = 1e-3;
                backproj = A{1}'*y{1};
                sigma2_init = var(backproj(find(abs(backproj) > ...
                    2*sqrt(sig2e_init))));
                ytrace = sum(diag(y{1}*y{1}'));
                Atrace = sum(diag(A{1}*A{1}'));
                lambda_init = min(max(0.01, (((ytrace - M*sig2e_init)/Atrace) - ...
                    abs(zeta_init)^2)/(sigma2_init - abs(zeta_init)^2)), 0.6);
                yxtrace = abs(sum(diag(y{1}*y{2}')));
                alpha_init = max(min(1 - yxtrace/(ytrace - M*sig2e_init), ...
                    0.5), 0.025);
                
                % Place the parameter initializations into the Param structure
                Params.lambda = lambda_init; 	Params.zeta = zeta_init;
                Params.sigma2 = sigma2_init;	Params.alpha = alpha_init;
                Params.sig2e = sig2e_init;
            end
            
            % Create a structure to hold the algorithm execution parameters
            RunOptions = Options();     % Start with default object
            RunOptions.inner_iters = inner_iters; 	RunOptions.alg = alg;
            RunOptions.smooth_iters = smooth_iters;	RunOptions.update = update;
            RunOptions.verbose = verbose;           RunOptions.eps = eps;

            if T == 1   % Need to use sp_frame_fxn only
                % Using perfect parameter knowledge
                Params.pi = lambda;     	Params.xi = GenParams.zeta;
                Params.psi = GenParams.sigma2;
                Params.sig2e = sig2e;       Params.eps = eps;
                EqOptions.iter = inner_iters; EqOptions.alg = 1;
                
                tic;
                X_mpi = cell(1,1);
                % Solve using sp_frame_fxn (single frame)
                [X_mpi{1}, OutMessages] = sp_frame_fxn(Y, Phi, Params, EqOptions);
                Run_MPI(a,b,n) = toc;
                
                lambda_hat = OutMessages.pi;
            else
                tic;
                % Solve using MPI-MMV
                [X_mpi, ~, lambda_hat] = sp_mmv_wrapper_fxn(y, A, Params, RunOptions);
                Run_MPI(a,b,n) = toc;
            end            
                        
            % Time-averaged normalized MSE (TNMSE)
            TNMSE_MPI(a,b,n) = sum(sum(abs(W_TRUE - [X_mpi{:}]).^2, 1)./sum(abs(W_TRUE).^2, 1))/T;
            
            % NSER
            % K-largest method
            RowNormSqs = sum(abs([X_mpi{:}]).^2, 2);
            [~, sortIndMPI] = sort(RowNormSqs, 1, 'descend');
            SuppEstMPI = sortIndMPI(1:K);
            NSER_MPI(a,b,n) = nser(true_support, SuppEstMPI);
            % Get a MAP estimate of support from lambda_hat
            NSER2_MPI(a,b,n) = nser(true_support, find(lambda_hat > 1/2));
            
        end     % lambda
    end     % alpha
    
    % Having completed one sweep of the (alpha, T) grid, estimate the
    % time remaining and save the temporary data
    est_time = (N_trials - n)*(etime(clock, time)/n)/60/60;
    disp(sprintf('(%d/%d) Estimated time remaining: %3.2f hours', ...
        n, N_trials, est_time));
    clear A Phi
    save(savefile);
end

disp(sprintf('Total time elapsed for test: %3.2f hours', ...
    etime(clock, time)/60/60));

try
    % Close the matlabpool
    matlabpool close
catch
    disp('No matlabpool to close')
end

% Final save of data
clear A Phi
save(savefile);