% PRODUCE_PLOTS     Produce plots for SWEEP_N_TEST
%

%
% Coded by: Justin Ziniel, The Ohio State Univ.
% E-mail: zinielj@ece.osu.edu
% Last change: 12/29/11
% Change summary: 
%		- Created from produce_plots v1.0 (12/29/11; JAZ)
% Version 1.1
%

clear; clc

%% Load a results file

uiload

%% Produce the plots

algs = {'T-MSBL', 'MSBL', 'SA-MUSIC', 'AMP-MMV', 'SKS'};    % Algorithms
suffix = {'tMSBL', 'MSBL', 'MUSIC', 'MPI', 'SKS'};
marker = {'-b*', '--bs', '-gd', '-ko', '--r+', '-k^'};

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Plot of the NMSE for each approach
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Compute the trial-averaged failure rate, MSE, Jaccard index, TNMSE, and 
% runtime statistics for each algorithm
for a = 1:N_alpha
    for d = 1:N_dim
        for g = 1:numel(algs)
            % First make sure that the current algorithm was included in
            % the dataset being plotted
            if exist(['FR_' suffix{g}]) == 1
                % Current algorithm is present in dataset

                % First identify those data points that are not NaNs, which may
                % be different for different algs. since MPI-MMV and SKS may be
                % averaged over more realizations
                eval(['inds = find(~isnan(TNMSE_' suffix{g} '(a,d,:)));']);

                % Compute the ensemble average over the non-NaN trials
                eval(['mean_FR_' suffix{g} '(a,d) = mean(FR_' suffix{g} ...
                    '(a,d,inds), 3);']);
                eval(['mean_MSE_' suffix{g} '(a,d) = mean(MSE_' suffix{g} ...
                    '(a,d,inds), 3);']);
                eval(['mean_NSER_' suffix{g} '(a,d) = mean(NSER_' suffix{g} ...
                    '(a,d,inds), 3);']);
                eval(['mean_TNMSE_' suffix{g} '(a,d) = mean(TNMSE_' suffix{g} ...
                    '(a,d,inds), 3);']);
                eval(['mean_Run_' suffix{g} '(a,d) = mean(Run_' suffix{g} ...
                    '(a,d,inds), 3);']);
                if strcmp('AMP-MMV', algs{g})
                    eval(['mean_NSER2_' suffix{g} '(a,d) = mean(NSER2_' suffix{g} ...
                        '(a,d,inds), 3);']);
                end
                % Keep track of the # of realizations averaged over
                eval(['N_trials_' suffix{g} '(a,d) = numel(inds);'])
                
            else
                % Current algorithm is missing, so just assign NaN
                eval(['mean_FR_' suffix{g} '(a,d) = NaN;']);
                eval(['mean_MSE_' suffix{g} '(a,d) = NaN;']);
                eval(['mean_NSER_' suffix{g} '(a,d) = NaN;']);
                eval(['mean_TNMSE_' suffix{g} '(a,d) = NaN;']);
                eval(['mean_Run_' suffix{g} '(a,d) = NaN;']);
                if strcmp('AMP-MMV', algs{g})
                    eval(['mean_NSER2_' suffix{g} '(a,d) = NaN;']);
                end
                % Keep track of the # of realizations averaged over
                eval(['N_trials_' suffix{g} '(a,d) = 0;'])
            end
        end
    end
end

% Plot the failure rate, MSE, and runtime for each algorithm at each
% choice of alpha
for a = 1:N_alpha
    % Clear existing figure
    figure(a); clf;
    
    subplot(1, 3, 2); hold on;  set(gca, 'XScale', 'log')
    for g = 1:numel(algs)
        plot(N_sweep, eval(['mean_FR_' suffix{g} '(a,:)']), marker{g})
    end
    hold off; grid on; axis square
    title(sprintf(['\\alpha = %1.2g  |  T = %d, M/N = %1.2g \\lambda = %1.3g SNR = %g dB'], ...
        alpha_sweep(a), T, delta, lambda, SNRdB));
    xlabel('Signal Dimension (N)'); ylabel('Failure Rate')
    legend(algs, 'Location', 'Best')
    
    subplot(1, 3, 1); hold on;  set(gca, 'XScale', 'log')
    for g = 1:numel(algs)
        semilogx(N_sweep, eval(['mean_MSE_' suffix{g} '(a,:)']), marker{g})
    end
    hold off; grid on; axis square
    title(sprintf(['\\alpha = %1.2g  |  T = %d, M/N = %1.2g \\lambda = %1.3g SNR = %g dB'], ...
        alpha_sweep(a), T, delta, lambda, SNRdB));
    xlabel('Signal Dimension (N)'); ylabel('Mean Square Error (MSE)')
    legend(algs, 'Location', 'Best')
    
    subplot(1, 3, 3); hold on; set(gca, 'XScale', 'log', 'YScale', 'log')
    for g = 1:numel(algs)
        loglog(N_sweep, eval(['mean_Run_' suffix{g} '(a,:)']), marker{g})
    end
    hold off; grid on; axis square
    title(sprintf(['\\alpha = %1.2g  |  T = %d, M/N = %1.2g \\lambda = %1.3g SNR = %g dB'], ...
        alpha_sweep(a), T, delta, lambda, SNRdB));
    xlabel('Signal Dimension (N)'); ylabel('Runtime (s)')
    legend(algs, 'Location', 'Best')
end


% Plot the Jaccard index, TNMSE, and runtime for each algorithm at each
% choice of alpha
for a = 1:N_alpha
    % Clear existing figure
    figure(a + N_alpha); clf;
    
    subplot(1, 3, 2); hold on; set(gca, 'XScale', 'log', 'YScale', 'log')
    for g = 1:numel(algs)
        if ~strcmp(suffix{g}, 'SKS')
            semilogx(N_sweep, eval(['mean_NSER_' suffix{g} '(a,:)']), marker{g})
        end
    end
    plot(N_sweep, mean_NSER2_MPI(a,:), marker{end});
    hold off; grid on; axis square
    title(sprintf(['\\alpha = %1.2g  |  T = %d, N/M = %1.2g, \\lambda = %1.3g, SNR = %g dB'], ...
        alpha_sweep(a), T, 1/delta, lambda, SNRdB)); axis tight
    xlabel('Signal Dimension (N)'); ylabel('Normalized Support Error Rate (NSER)')
    legend({algs{1:4}, 'AMP-MMV [p(s_n|{\bf y})]'}, 'Location', 'Best')
    
    subplot(1, 3, 1); hold on; set(gca, 'XScale', 'log')
    for g = 1:numel(algs)
        semilogx(N_sweep, 10*log10(eval(['mean_TNMSE_' suffix{g} '(a,:)'])), marker{g})
    end
    hold off; grid on; axis square
    title(sprintf(['\\alpha = %1.2g  |  T = %d, N/M = %1.2g, \\lambda = %1.3g, SNR = %g dB'], ...
        alpha_sweep(a), T, 1/delta, lambda, SNRdB)); axis tight
    xlabel('Signal Dimension (N)'); ylabel('Timestep-Averaged Normalized MSE (TNMSE) [dB]')
    legend(algs, 'Location', 'Best')
    
    subplot(1, 3, 3); hold on; set(gca, 'XScale', 'log', 'YScale', 'log')
    for g = 1:numel(algs)
        if ~strcmp(suffix{g}, 'SKS')
            loglog(N_sweep, eval(['mean_Run_' suffix{g} '(a,:)']), marker{g})
        end
    end
    hold off; grid on; axis square
    title(sprintf(['\\alpha = %1.2g  |  T = %d, N/M = %1.2g, \\lambda = %1.3g, SNR = %g dB'], ...
        alpha_sweep(a), T, 1/delta, lambda, SNRdB)); axis tight
    xlabel('Signal Dimension (N)'); ylabel('Runtime [s]')
    legend(algs, 'Location', 'Best')
end
