% SWEEP_N_TEST 	A test of AMP-MMV, and the support-aware Kalman smoother 
% (SKS), as well as the algorithms T-MSBL, and MSBL of Zhilin
% Zhang and Bhaskar Rao's 2011 paper "Sparse Signal Recovery with
% Temporally Correlated Source Vectors Using Sparse Bayesian Learning".  
% The test will fix all signal model parameters and observe how performance
% changes as problem dimensions grow.  Specifically, the number of
% unknowns, N, will be swept over a range, with M growing proportionally,
% i.e., M = delta*N, for delta in [0,1].
%
% Notation will be consistent with the AMP-MMV signal model.
%
%
% SYNTAX:
% sweep_N_test(T, delta, lambda, SNRdB, N_trials)
%
% INPUTS:
%   T               Number of timesteps
%   delta           Ratio of measurements/unknowns
%   lambda          Activity probability, i.e., Pr{s_n = 1} = lambda
%   SNRdB           SNR at each timestep (in dB)
%   N_trials     	Number of trials to average results over
%
% OUTPUTS:
% This function will save the MSEs (||X - X_hat||_F^2 / ||X||_F^2, where
% ||B||_F^2 is the square of the Frobenius norm of the matrix B) and 
% failure rates (defined in ref. [28] of Zhang and Rao's paper) for the
% T-MSBL, and MSBL algorithms, as well as for AMP-MMV, a 
% support-aware Kalman smoother, and Lee and Bresler's Subspace-Augmented
% MUSIC algorithm.  Slightly different performance metrics are also 
% computed and saved.  These include timestep-averaged normalized MSE
% (TNMSE) and the normalized support error rate (NSER), measuring the
% similarity between the estimated and true support set.  Results are 
% stored in a .MAT file whose filename is created at runtime based on input
% parameters and the time/date of the start of execution.
% Suffix keywords:
%   _tMSBL      The T-MSBL algorithm
%   _MSBL       The MSBL algorithm
%   _MPI        The multi-timestep BP algorithm (AMP-MMV)
%   _SKS        The support-aware Kalman smoother
%   _MUSIC      The subspace-augmented MUSIC algorithm (SA-MUSIC)
%

%
% Coded by: Justin Ziniel, The Ohio State Univ.
% E-mail: zinielj@ece.osu.edu
% Last change: 12/29/11
% Change summary: 
%		- Created from sweep_N_test v1.0 (12/29/11; JAZ)
% Version 1.1
%


function sweep_N_test(T, delta, lambda, SNRdB, N_trials)

%% Declare the parameters for the signal model and test setup

if nargin == 0
    T = 4;          	% Number of timesteps
    delta = 1/3;      	% Ratio of measurements/unknowns
    lambda = 0.15;    	% Activity probability, i.e., Pr{s_n = 1} = lambda
    SNRdB = 25;       	% SNR at each timestep (in dB)
    N_trials = 250;   	% Number of trials to average results over
end

warning('SKS is being run using AMP instead of ABP')

% Test setup parameters
alpha_sweep = [0.1, 0.05];              % Correlation coefficients to use
N_sweep = round(logspace(2, 4, 5)); 	% Problem dimensions to sweep over
N_alpha = numel(alpha_sweep);
N_dim = numel(N_sweep);

% Important BP execution parameters
inner_iters = 15;	% Max # of inner AMP iterations
smooth_iters = 5;  	% Max # of smoothing iterations
alg = 'AMP';        % Use AMP, not BP
update = true;    	% Use EM parameter estimation
verbose = true;    	% SP_MMV_WRAPPER_FXN verbosity
eps = 1e-7;         % Regularization parameter

% Filename for saving data
savefile = [datestr(now, 'ddmmmyyyy_HHMMSSFFF') '_sweep_N_test_T_' ...
    num2str(T) '_delta_' num2str(delta) '_lambda_' num2str(lambda) '_SNRdB_' ...
    num2str(SNRdB) '.mat'];

% Randomly seed the RNG and save its state for reference purposes
NewStream = RandStream.create('mt19937ar', 'Seed', sum(100*clock));
RandStream.setDefaultStream(NewStream);
savedState = NewStream.State;


%% Execute the sweep N test

% Create arrays to store the failure rates and MSEs
FR_tMSBL = NaN*ones(N_alpha, N_dim, N_trials);   	% Failure rate for T-MSBL
FR_MSBL = NaN*ones(N_alpha, N_dim, N_trials);    	% Failure rate for MSBL
FR_MPI = NaN*ones(N_alpha, N_dim, N_trials);      	% Failure rate for MPI-MMV
FR_SKS = NaN*ones(N_alpha, N_dim, N_trials);      	% Failure rate for SKS
FR_MUSIC = NaN*ones(N_alpha, N_dim, N_trials);      % Failure rate for SA-MUSIC
MSE_tMSBL = NaN*ones(N_alpha, N_dim, N_trials);   	% MSE for T-MSBL
MSE_MSBL = NaN*ones(N_alpha, N_dim, N_trials);     	% MSE for MSBL
MSE_MPI = NaN*ones(N_alpha, N_dim, N_trials);      	% MSE for MPI-MMV
MSE_SKS = NaN*ones(N_alpha, N_dim, N_trials);      	% MSE for SKS
MSE_MUSIC = NaN*ones(N_alpha, N_dim, N_trials);   	% MSE for SA-MUSIC
NSER_tMSBL = NaN*ones(N_alpha, N_dim, N_trials);    % NSER for T-MSBL
NSER_MSBL = NaN*ones(N_alpha, N_dim, N_trials);    	% NSER for MSBL
NSER_MPI = NaN*ones(N_alpha, N_dim, N_trials);    	% NSER for MPI-MMV
NSER_SKS = NaN*ones(N_alpha, N_dim, N_trials);    	% NSER for SKS
NSER_MUSIC = NaN*ones(N_alpha, N_dim, N_trials);   	% NSER for SA-MUSIC
NSER2_MPI = NaN*ones(N_alpha, N_dim, N_trials);    	% 2nd NSER for MPI-MMV
TNMSE_tMSBL = NaN*ones(N_alpha, N_dim, N_trials);  	% TNMSE for T-MSBL
TNMSE_MSBL = NaN*ones(N_alpha, N_dim, N_trials);   	% TNMSE for MSBL
TNMSE_MPI = NaN*ones(N_alpha, N_dim, N_trials);    	% TNMSE for MPI-MMV
TNMSE_SKS = NaN*ones(N_alpha, N_dim, N_trials);   	% TNMSE for SKS
TNMSE_MUSIC = NaN*ones(N_alpha, N_dim, N_trials);	% TNMSE for SA-MUSIC
Run_tMSBL = NaN*ones(N_alpha, N_dim, N_trials);    	% Runtime for T-MSBL
Run_MSBL = NaN*ones(N_alpha, N_dim, N_trials);     	% Runtime for MSBL
Run_MPI = NaN*ones(N_alpha, N_dim, N_trials);      	% Runtime for MPI-MMV
Run_SKS = NaN*ones(N_alpha, N_dim, N_trials);      	% Runtime for SKS
Run_MUSIC = NaN*ones(N_alpha, N_dim, N_trials);   	% Runtime for SA-MUSIC

% try % Matlabpool opening
%     % Open a matlabpool, if one is not already open
%     if matlabpool('size') == 0
%         matlabpool('open', min(N_dim, 8));
%         disp(sprintf('Opened a matlabpool with %d workers', matlabpool('size')));
%     end
% catch
%     disp('Error in opening a matlabpool')
% end

time = clock;   % Start the stopwatch
for n = 1:N_trials
    for a = 1:N_alpha     % Chosen alpha
        alpha = alpha_sweep(a);
        for d = 1:N_dim  	% Number of unknowns
            N = N_sweep(d);
            M = round(delta*N);
            
            % *************************************************************
            % Generate a signal/measurement pair realization for this suite
            % according to the MPI-MMV signal model
            % *************************************************************
            % Construct the signal generation parameter object
            GenParams = SigGenParams();  % Start w/ default object
            GenParams.N = N;             GenParams.M = M;
            GenParams.T = T;             GenParams.lambda = lambda;   
            GenParams.zeta = 0;          GenParams.sigma2 = 1;        
            GenParams.alpha = alpha;     GenParams.SNRmdB = SNRdB;    
            GenParams.version = 'R';     GenParams.A_type = 1;

            % ***********************************************************
            % Generate the time series according to the above parameters
            [x_true, y, A, true_support, K, sig2e] = signal_gen_fxn(GenParams);

            % Create variables based on the realization that better match
            % the signal model of T-MSBL, T-SBL, and MSBL
            W_TRUE = [x_true{:}];
            Y = [y{:}];
            Phi = A{1};
            stdnoise = sqrt(sig2e);


            % *************************************************************
            %                   Solve using T-MSBL
            % *************************************************************
            tic;
            % Run the T-MSBL algorithm.  Use default values for all other
            % parameters (see T-MSBL code) (Change to PRUNE_GAMMA was
            % hard-coded into TMSBL.m (1e-2))
            if SNRdB > 22
                [Weight1] = TMSBL(Phi, Y, 'noise', 'small');
            elseif 6 < SNRdB && SNRdB < 22
                [Weight1] = TMSBL(Phi, Y, 'noise', 'mild');
            else
                [Weight1] = TMSBL(Phi, Y, 'noise', 'large');
            end
            Run_tMSBL(a,d,n) = toc;

            % If perfSupp(.) ~= 1, then imperfect support est. (failure)
            FR_tMSBL(a,d,n) = perfSupp(Weight1, true_support, 'firstlargest', K) ~= 1;               

            % Mean Square Error (MSE)
            MSE_tMSBL(a,d,n) = (norm(W_TRUE - Weight1,'fro') / norm(W_TRUE,'fro'))^2;

            % Time-averaged normalized MSE (TNMSE)
            TNMSE_tMSBL(a,d,n) = sum(sum((W_TRUE - Weight1).^2, 1)./sum(W_TRUE.^2, 1))/T;

            % NSER
            % Here we have some flexibility in defining the estimated
            % support obtained using T-MSBL.  We could use all the indices
            % contained in gamma_ind1, however this would include many
            % negligible coefficients.  A better estimate, although
            % biased in favor of T-MSBL, is to assume that the only "true"
            % elements of the support correspond to the K largest magnitude
            % rows of Weight1.  This is the method we will use.
            RowNormSqs = sum(Weight1.^2,2);
            [~, sortInd1] = sort(RowNormSqs, 1, 'descend');
            SuppEst1 = sortInd1(1:K);
            NSER_tMSBL(a,d,n) = nser(true_support, SuppEst1);
            
            
            % *************************************************************
            %                   Solve using MSBL
            % *************************************************************
            % ** Note that 'Lambda' here is unrelated to 'lambda'
            Lambda = 1e-3;      % Initial choice of regularization param. 
            Learn_Lambda = 1;   % Learn lambda from data.  Not robust for
                                % SNR < 20dB
            PRUNE_GAMMA = 1e-2;	% threshold for prunning small hyperparameters gamma_i
            EPSILON     = 1e-8;	% threshold for stopping iteration. 
            MAX_ITERS   = 2000;	% maximum iterations
            
            tic;
            % Run the MSBL algorithm.  Use default values for all other
            % parameters (see MSBL code)
            [Weight3] = MSBL(Phi, Y, Lambda, Learn_Lambda, ...
                    'prune_gamma', PRUNE_GAMMA, 'max_iters', MAX_ITERS, ...
                    'epsilon', EPSILON, 'print', 0);
            Run_MSBL(a,d,n) = toc;

            % If perfSupp(.) ~= 1, then imperfect support est. (failure)
            FR_MSBL(a,d,n) = perfSupp(Weight3, true_support, 'firstlargest', K) ~= 1;               

            % Mean Square Error (MSE)
            MSE_MSBL(a,d,n) = (norm(W_TRUE - Weight3,'fro') / norm(W_TRUE,'fro'))^2;

            % Time-averaged normalized MSE (TNMSE)
            TNMSE_MSBL(a,d,n) = sum(sum((W_TRUE - Weight3).^2, 1)./sum(W_TRUE.^2, 1))/T;

            % NSER
            % Use the K-largest method, like T-MSBL
            RowNormSqs = sum(Weight3.^2,2);
            [~, sortInd3] = sort(RowNormSqs, 1, 'descend');
            SuppEst3 = sortInd3(1:K);
            NSER_MSBL(a,d,n) = nser(true_support, SuppEst3);
            
            
            % *************************************************************
            %       Solve using the support-aware Kalman smoother
            % *************************************************************            
            % Create the signal model parameter object
            Params = ModelParams(GenParams, sig2e);

            % Create an object to hold SKS runtime options
            RunOptions = Options();     %Start with default object
            RunOptions.inner_iters = inner_iters;
            RunOptions.smooth_iters = smooth_iters;
            % (Defaults to ABP mode, not AMP mode)
            RunOptions.alg = alg;      RunOptions.eps = eps;
            
            if T == 1   % Need to use sp_frame_fxn only
                Params.pi = zeros(N,1);     Params.pi(true_support) = 1;
                Params.xi = GenParams.zeta;	
                Params.psi = GenParams.sigma2;
                Params.eps = eps;
                EqOptions.iter = inner_iters;  EqOptions.alg = 2;
                
                tic;
                X_sks = cell(1,1);
                % Solve using sp_frame_fxn (single frame)
                X_sks{1} = sp_frame_fxn(Y, Phi, Params, EqOptions);
                Run_SKS(a,d,n) = toc;
            else
                tic;
                % Solve using the SKS
                [X_sks] = genie_multi_frame_fxn(y, A, true_support, Params, RunOptions);
                Run_SKS(a,d,n) = toc;
            end

            % If perfSupp(.) ~= 1, then imperfect support est. (failure)
            FR_SKS(a,d,n) = perfSupp([X_sks{:}], true_support, 'firstlargest', K) ~= 1;               

            % Mean Square Error (MSE)
            MSE_SKS(a,d,n) = (norm(W_TRUE - [X_sks{:}],'fro') / norm(W_TRUE,'fro'))^2;
            
            % Time-averaged normalized MSE (TNMSE)
            TNMSE_SKS(a,d,n) = sum(sum((W_TRUE - [X_sks{:}]).^2, 1)./sum(W_TRUE.^2, 1))/T;
            
            % NSER
            NSER_SKS(a,d,n) = nser(true_support, ...
                find(sum(abs([X_sks{:}]),2) ~= 0));
            
            
            % *************************************************************
            %                       Solve using MPI-MMV
            % *************************************************************
            % Create an object to hold the signal model parameters
            Params = ModelParams(GenParams, sig2e);
            
            % Try to guess some decent initial values for the parameters
            if T > 1
                zeta_init = 0;
                sig2e_init = 1e-3;
                backproj = A{1}'*y{1};
                sigma2_init = var(backproj(find(abs(backproj) > ...
                    2*sqrt(sig2e_init))));
                ytrace = sum(diag(y{1}*y{1}'));
                Atrace = sum(diag(A{1}*A{1}'));
                lambda_init = min(max(0.01, (((ytrace - M*sig2e_init)/Atrace) - ...
                    abs(zeta_init)^2)/(sigma2_init - abs(zeta_init)^2)), 0.6);
                yxtrace = abs(sum(diag(y{1}*y{2}')));
                alpha_init = max(min(1 - yxtrace/(ytrace - M*sig2e_init), ...
                    0.5), 0.025);
                
                % Place the parameter initializations into the Param structure
                Params.lambda = lambda_init; 	Params.zeta = zeta_init;
                Params.sigma2 = sigma2_init;	Params.alpha = alpha_init;
                Params.sig2e = sig2e_init;
            end
            
            % Create a structure to hold the algorithm execution parameters
            RunOptions = Options();     % Start with default object
            RunOptions.inner_iters = inner_iters; 	RunOptions.alg = alg;
            RunOptions.smooth_iters = smooth_iters;	RunOptions.update = update;
            RunOptions.verbose = verbose;           RunOptions.eps = eps;

            if T == 1   % Need to use sp_frame_fxn only
                % Using perfect parameter knowledge
                Params.pi = lambda;     	Params.xi = GenParams.zeta;
                Params.psi = GenParams.sigma2;
                Params.sig2e = sig2e;       Params.eps = eps;
                EqOptions.iter = inner_iters; EqOptions.alg = 1;
                
                tic;
                X_mpi = cell(1,1);
                % Solve using sp_frame_fxn (single frame)
                [X_mpi{1}, OutMessages] = sp_frame_fxn(Y, Phi, Params, EqOptions);
                Run_MPI(a,d,n) = toc;
                
                lambda_hat = OutMessages.pi;
            else
                tic;
                % Solve using AMP-MMV
                [X_mpi, ~, lambda_hat] = sp_mmv_wrapper_fxn(y, A, Params, RunOptions);
                Run_MPI(a,d,n) = toc;
            end

            % If perfSupp(.) ~= 1, then imperfect support est. (failure)
            FR_MPI(a,d,n) = perfSupp([X_mpi{:}], true_support, 'firstlargest', K) ~= 1;               

            % Mean Square Error (MSE)
            MSE_MPI(a,d,n) = (norm(W_TRUE - [X_mpi{:}],'fro') / norm(W_TRUE,'fro'))^2;
            
                        
            % Time-averaged normalized MSE (TNMSE)
            TNMSE_MPI(a,d,n) = sum(sum((W_TRUE - [X_mpi{:}]).^2, 1)./sum(W_TRUE.^2, 1))/T;
            
            % NSER
            % For fairness, we will also use the K-largest row-norm method,
            % but will also use a more realistic support estimator obtained
            % from posteriors on the binary support variables
            % K-largest method
            RowNormSqs = sum([X_mpi{:}].^2, 2);
            [~, sortIndMPI] = sort(RowNormSqs, 1, 'descend');
            SuppEstMPI = sortIndMPI(1:K);
            NSER_MPI(a,d,n) = nser(true_support, SuppEstMPI);
            % Get a MAP estimate of support from lambda_hat
            NSER2_MPI(a,d,n) = nser(true_support, find(lambda_hat > 1/2));
            
            
            % *************************************************************
            %                   Solve using SA-MUSIC
            % *************************************************************    
            tic;
            % Run the SA-MUSIC algorithm.  Assume perfect knowledge of the
            % true number of non-zero ceofficients, K
            [SuppEstMUSIC] = SAMUSIC(Y, Phi, K);
            Run_MUSIC(a,d,n) = toc;
            
            % Use the support estimate obtained from SA-MUSIC to obtain an
            % MMSE estimate of X, conditioned on the support.  Allow for
            % perfect parameter knowledge
            K_hat = numel(SuppEstMUSIC);
            CovMat = SigGenParams.kappa_0 * Phi(:,SuppEstMUSIC) * ...
                Phi(:,SuppEstMUSIC)' + sig2e*eye(M);
            X_music = zeros(N,T);
            X_music(SuppEstMUSIC, :) = SigGenParams.eta_0*ones(K_hat,T) + ...
                SigGenParams.kappa_0*Phi(:,SuppEstMUSIC)'*(CovMat\(Y - ...
                SigGenParams.eta_0*repmat(sum(Phi(:,SuppEstMUSIC), 2), 1, T)));

            % If perfSupp(.) ~= 1, then imperfect support est. (failure)
            FR_MUSIC(a,d,n) = perfSupp(X_music, true_support, 'firstlargest', K) ~= 1;               

            % Mean Square Error (MSE)
            MSE_MUSIC(a,d,n) = (norm(W_TRUE - X_music,'fro') / norm(W_TRUE,'fro'))^2;

            % Time-averaged normalized MSE (TNMSE)
            TNMSE_MUSIC(a,d,n) = sum(sum((W_TRUE - X_music).^2, 1)./sum(W_TRUE.^2, 1))/T;
            
            % NSER
            % Use the returned support estimate, J_hat, which should be
            % equivalent to the K-largest-row-norm method, since SA-MUSIC
            % is returning a solution with exactly K non-zero rows
            NSER_MUSIC(a,d,n) = nser(true_support, SuppEstMUSIC);
            
            clear CovMat
            
        end     % N
    end     % alpha
    
    % Having completed one sweep of the (alpha, T) grid, estimate the
    % time remaining and save the temporary data
    est_time = (N_trials - n)*(etime(clock, time)/n)/60/60;
    disp(sprintf('(%d/%d) Estimated time remaining: %3.2f hours', ...
        n, N_trials, est_time));
    clear A Phi
    save(savefile);
end

disp(sprintf('Total time elapsed for test: %3.2f hours', ...
    etime(clock, time)/60/60));

try
    % Close the matlabpool
    matlabpool close
catch
    disp('No matlabpool to close')
end

% Final save of data
clear A Phi
save(savefile);
